local Addon, private = ...

-- Builtins
local _G = _G
local math = math
local os = os
local pairs = pairs
local pcall = pcall
local print = print
local select = select
local string = string
local table = table
local tonumber = tonumber
local tostring = tostring
local unpack = unpack

-- Globals
local Command = Command
local dump = dump
local Event = Event
local Inspect = Inspect
local UI = UI
local UIParent = UIParent
local Utility = Utility
local zlib = zlib

-- Locals
local padding = 5
local identifierWidth = 90
local scrollbarWidth = 16
local playerName = Inspect.Unit.Detail("player").name
local incomingColor = { 0, 1, 1 }
local outgoingColor = { 1, 1, 0}
local labelFormats = {
	player = "Player Storage:\n(%i/%i used)",
	guild = "Guild Storage:\n(%i/%i used)",
	size = "Size: %i",
	checksum = "Checksum: %s",
	uncompressed = "Uncompressed: %i",
	compressed = "Compressed: %i",
	ratio = "Ratio: %i%%",
}
local messageHistory = {
--[[
	-- List of messages in chronological order: [1] = oldest, [#] = newest
	[*] = {
		from = string, -- "player" if send/broadcast, otherwise receive
		type = string, -- "tell", "channel", "guild", "officer", "party", "raid", "say", "yell", "send"
		channel = string/nil,
		identifier = string,
		data = string,
		timestamp = number, -- from os.time()
	},
]]
}
local statistics = {
--[[
	[identifier] = {
		[type] = {
			sent = number,
			received = number,
		},
	},
]]
}

local typeShortcuts = {
	tell = "tell",
	channel = "chan",
	guild = "guild",
	officer = "offi",
	party = "party",
	raid = "raid",
	say = "say",
	yell = "yell",
	send = "send",
	["*"] = "*",
}
local typeColors = {
	tell = { 251 / 255, 142 / 255, 251 / 255 },
	channel = { 251 / 255, 186 / 255, 189 / 255 },
	guild = { 63 / 255, 248 / 255, 63 / 255 },
	officer = { 53 / 255, 184 / 255, 53 / 255 },
	party = { 168 / 255, 165 / 255, 251 / 255 },
	raid = { 252 / 255, 124 / 255, 1 / 255 },
	say = { 251 / 255, 251 / 255, 251 / 255 },
	yell = { 251 / 255, 62 / 255, 63 / 255 },
	send = { 251 / 255, 142 / 255, 251 / 255 },
	["*"] = { 251 / 255, 251 / 255, 0 / 255 },
}

local oldCommandMessageAccept = Command.Message.Accept
local oldCommandMessageBroadcast = Command.Message.Broadcast
local oldCommandMessageReject = Command.Message.Reject
local oldCommandMessageSend = Command.Message.Send

-- Prototypes
local CommandMessageAccept
local CommandMessageBroadcast
local CommandMessageReject
local CommandMessageSend
local acceptorSelected
local statisticSelected
local messageSelected
local getHistory
local resetCounters
local clearEntireHistory
local clearDisplayedHistory
local simulateIncomingBroadcast
local simulateIncomingSend

setfenv(1, private)

if(Addon.toc.debug) then
	log = print
else
	log = function() end
end

-- UI creation methods
-- ============================================================================

local function createTextField(parent, defaultText)
	local background = UI.CreateFrame("Frame", "", parent)
	local self = UI.CreateFrame("RiftTextfield", "", background)
	self:SetBackgroundColor(0, 0, 0, 0.7)
	
	background:SetBackgroundColor(1, 1, 1, 0.8)
	background:SetPoint("TOPLEFT", self, "TOPLEFT", -1, -1)
	background:SetPoint("BOTTOMRIGHT", self, "BOTTOMRIGHT", 1, 1)
	
	if(defaultText) then
		function self.Event:KeyFocusGain()
			if(self:GetText() == defaultText) then
				self:SetText("")
			end
		end

		function self.Event:KeyFocusLoss()
			if(self:GetText() == "") then
				self:SetText(defaultText)
			end
		end
		self:SetText(defaultText)
	else
		self:SetText("")
	end

	return self
end

local function createDataTextField(parent)
	local self = createTextField(parent)

	function self.Event:KeyDown(button)
		local code = string.byte(button)
		local txt = self:GetText()
		local pos = self:GetCursor()
		self:SetKeyFocus(true) 
		if tonumber(code) == 13 then
			self:SetText(string.insert(txt, pos + 1, "\n"))
			self:SetCursor(pos + 1)
		elseif tonumber(code) == 9 then
			self:SetText(string.insert(txt, pos + 1, "\t"))
			self:SetCursor(pos + 1)
		end
	end
--[[
	function self.Event:TextfieldChange()
		parent.realText = self:GetText()
		dataOrIdentifierChanged(parent, self)
	end
]]
	return self
end

local function createCheckbox(parent, text, checked)
	local self = UI.CreateFrame("RiftCheckbox", "", parent)
	self.label = UI.CreateFrame("Text", "", self)
	self.label:SetPoint("LEFTCENTER", self, "RIGHTCENTER")
	self.label:SetText(text)
	self:SetChecked(checked)
	
	function self.label.Event.LeftDown()
		self:SetChecked(not self:GetChecked())
	end
	
	return self
end

local function createAcceptorsList(parent, width, entries)
	local self = UI.CreateFrame("Frame", "", parent)
	self:SetBackgroundColor(0, 0, 0)
	self.selected = 0
	
	local scrollbar = UI.CreateFrame("RiftScrollbar", "", self)
	scrollbar:SetPoint("TOPLEFT", self, "TOPRIGHT")
	scrollbar:SetPoint("BOTTOMLEFT", self, "BOTTOMRIGHT")
	scrollbar.Event.ScrollbarChange = function() self:Update() end
	
	local prevAnchor = self
	self.entries = { }
	self.identifiers = { }
	self.types = { }
	self.accepted = { }
	
	for i = 1, entries do
		local entry = UI.CreateFrame("Frame", "", self)
		entry:SetPoint("TOPLEFT", prevAnchor, prevAnchor == self and "TOPLEFT" or "BOTTOMLEFT")
		entry:SetPoint("TOPRIGHT", prevAnchor, prevAnchor == self and "TOPRIGHT" or "BOTTOMRIGHT")
		
		local reject = UI.CreateFrame("RiftButton", "", entry)
		reject:SetSkin("close")
		reject:SetPoint("TOPLEFT", entry, "TOPLEFT")
		reject:SetPoint("BOTTOMLEFT", entry, "BOTTOMLEFT")
		reject.Event.LeftPress = function()
			local acceptor = self.acceptors[entry.index + math.floor(scrollbar:GetPosition())]
			local identifier = acceptor[1]
			local type = acceptor[2]
			if(identifier == "*") then
				identifier = nil
			end
			if(type == "*") then
				type = nil
			end
			Command.Message.Reject(type, identifier)
		end
		
		local accepted = UI.CreateFrame("Text", "", entry)
		accepted:SetPoint("TOPRIGHT", entry, "TOPRIGHT")
		accepted:SetText("0")
		
		local type = UI.CreateFrame("Text", "", entry)
		type:SetText("X")
		type:SetPoint("TOPRIGHT", entry, "TOPRIGHT", -30, 0)
		type:SetWidth(30)
		
		local identifier = UI.CreateFrame("Text", "", entry)
		identifier:SetText("X")
		identifier:SetPoint("TOPLEFT", reject, "TOPRIGHT")
		identifier:SetPoint("TOPRIGHT", type, "TOPLEFT")
		
		entry:SetHeight(identifier:GetHeight())
		reject:SetWidth(reject:GetHeight())
		
		entry.Event.MouseIn = function(entry) entry:SetBackgroundColor(0.3, 0.3, 0.3) end
		entry.Event.MouseOut = function(entry)
			if(i + math.floor(scrollbar:GetPosition() + 0.5) ~= self.selected) then
				entry:SetBackgroundColor(0, 0, 0)
			end
		end
		entry.Event.LeftDown = function(entry)
			self.selected = i + math.floor(scrollbar:GetPosition() + 0.5)
			acceptorSelected(parent, self, self.selected)
			self:Update()
		end
		table.insert(self.entries, entry)
		table.insert(self.identifiers, identifier)
		table.insert(self.types, type)
		table.insert(self.accepted, accepted)
		prevAnchor = entry
	end
	
	self:SetHeight(self.entries[1]:GetHeight() * entries)
	self:SetWidth(width)
	
	function self:SetAcceptors(acceptors)
		self.acceptors = { }
		for k, v in pairs(acceptors) do
			self.acceptors[#self.acceptors + 1] = { k[2] or "*", k[1] or "*", tostring(v) }
		end
		table.sort(self.acceptors, function(a, b) return a[1] < b[1] end)
		scrollbar:SetRange(0, math.max(1, #self.acceptors - entries))
		scrollbar:SetEnabled(#self.acceptors > entries)
		scrollbar:SetPosition(0)
		self:Update()
	end
	
	function self:Update()
		local offset = math.floor(scrollbar:GetPosition() + 0.5)
		for i = 1, entries do
			local index = i + offset
			if(index > #self.acceptors) then
				self.entries[i]:SetVisible(false)
			else
				self.entries[i]:SetVisible(true)
				self.identifiers[i]:SetText(self.acceptors[index][1])
				self.types[i]:SetText(typeShortcuts[self.acceptors[index][2]])
				self.types[i]:SetFontColor(unpack(typeColors[self.acceptors[index][2]]))
				self.accepted[i]:SetText(self.acceptors[index][3])
				if(index == self.selected) then
					self.entries[i]:SetBackgroundColor(0.3, 0.3, 0.3)
				else
					self.entries[i]:SetBackgroundColor(0, 0, 0)
				end
			end
		end
	end
	
	return self
end

local function createHistoryList(parent, entries)
	local self = UI.CreateFrame("Frame", "", parent)
	self:SetBackgroundColor(0, 0, 0)
	self.selected = 0
	
	local scrollbar = UI.CreateFrame("RiftScrollbar", "", self)
	scrollbar:SetPoint("TOPLEFT", self, "TOPRIGHT")
	scrollbar:SetPoint("BOTTOMLEFT", self, "BOTTOMRIGHT")
	scrollbar.Event.ScrollbarChange = function() self:Update() end
	
	local prevAnchor = self
	self.entries = { }
	self.identifiers = { }
	self.froms = { }
	self.types = { }
	self.targets = { }
	self.timestamps = { }
	
	self.history = { }
	
	for i = 1, entries do
		local entry = UI.CreateFrame("Frame", "", self)
		entry:SetPoint("TOPLEFT", prevAnchor, prevAnchor == self and "TOPLEFT" or "BOTTOMLEFT")
		entry:SetPoint("TOPRIGHT", prevAnchor, prevAnchor == self and "TOPRIGHT" or "BOTTOMRIGHT")
		
		local identifier = UI.CreateFrame("Text", "", entry)
		identifier:SetText("SomeIdentifier")
		identifier:SetPoint("TOPLEFT", entry, "TOPLEFT")
		identifier:SetWidth(identifierWidth)
		
		local from = UI.CreateFrame("Text", "", entry)
		from:SetText("Someplayer")
		from:SetPoint("TOPLEFT", identifier, "TOPRIGHT")
		from:SetWidth(80)
		
		local target = UI.CreateFrame("Text", "", entry)
		target:SetText("Someplayer")
		target:SetPoint("TOPLEFT", from, "TOPRIGHT")
		target:SetWidth(80)
		
		local type = UI.CreateFrame("Text", "", entry)
		type:SetText("chan")
		type:SetPoint("TOPLEFT", target, "TOPRIGHT")
		type:SetWidth(30)
		
		local timestamp = UI.CreateFrame("Text", "", entry)
		timestamp:SetText("")
		timestamp:SetPoint("TOPRIGHT", entry, "TOPRIGHT")
		timestamp:SetWidth(60)
		
		entry:SetHeight(identifier:GetHeight())
		
		entry.Event.MouseIn = function(entry) entry:SetBackgroundColor(0.3, 0.3, 0.3) end
		entry.Event.MouseOut = function(entry)
			if(i + math.floor(scrollbar:GetPosition() + 0.5) ~= self.selected) then
				entry:SetBackgroundColor(0, 0, 0)
			end
		end
		entry.Event.LeftDown = function(entry)
			self.selected = i + math.floor(scrollbar:GetPosition() + 0.5)
			messageSelected(parent, self, self.history[i + math.floor(scrollbar:GetPosition() + 0.5)])
			self:Update()
		end
		table.insert(self.entries, entry)
		table.insert(self.identifiers, identifier)
		table.insert(self.froms, from)
		table.insert(self.types, type)
		table.insert(self.targets, target)
		table.insert(self.timestamps, timestamp)
		prevAnchor = entry
	end
	
	self:SetHeight(self.entries[1]:GetHeight() * entries)
	self:SetWidth(identifierWidth + 80 + 30 + 80 + 60)
	
	local label = UI.CreateFrame("Text", "", self)
	label:SetText("from")
	label:SetPoint("BOTTOMLEFT", self, "TOPLEFT", identifierWidth, 3)
	
	local label = UI.CreateFrame("Text", "", self)
	label:SetText("target")
	label:SetPoint("BOTTOMLEFT", self, "TOPLEFT", identifierWidth + 80, 3)
	
	local label = UI.CreateFrame("Text", "", self)
	label:SetText("type")
	label:SetPoint("BOTTOMLEFT", self, "TOPLEFT", identifierWidth + 80 + 80, 3)
	
	function self:SelectHistory(identifier, type)
		self.identifier = identifier
		self.type = type
		self.selected = 0
		messageSelected(parent, self, nil)
		self:UpdateHistory()
		scrollbar:SetPosition(0)
	end
	
	function self:UpdateHistory()
		local min, max = scrollbar:GetRange()
		local pos = math.floor(scrollbar:GetPosition())
		local bottom = max == 1 or pos == max
		self.history = getHistory(self.identifier, self.type)

		scrollbar:SetRange(0, math.max(1, #self.history - entries))
		scrollbar:SetEnabled(#self.history > entries)
		if(bottom) then
			min, max = scrollbar:GetRange()
			scrollbar:SetPosition(max == 1 and 0 or max)
		end
		self:Update()
	end
	
	function self:Update()
		local offset = math.floor(scrollbar:GetPosition() + 0.5)
		for i = 1, entries do
			local index = i + offset
			if(index > #self.history) then
				self.entries[i]:SetVisible(false)
			else
				self.entries[i]:SetVisible(true)
				local from = self.history[index].from
				if(from == "player") then
					self.froms[i]:SetText(playerName)
					self.identifiers[i]:SetFontColor(unpack(outgoingColor))
					self.froms[i]:SetFontColor(unpack(outgoingColor))
					self.targets[i]:SetFontColor(unpack(outgoingColor))
					self.timestamps[i]:SetFontColor(unpack(outgoingColor))
				else
					self.froms[i]:SetText(from)
					self.identifiers[i]:SetFontColor(unpack(incomingColor))
					self.froms[i]:SetFontColor(unpack(incomingColor))
					self.targets[i]:SetFontColor(unpack(incomingColor))
					self.timestamps[i]:SetFontColor(unpack(incomingColor))
				end
				self.identifiers[i]:SetText(self.history[index].identifier)
				self.types[i]:SetFontColor(unpack(typeColors[self.history[index].type]))
				self.types[i]:SetText(typeShortcuts[self.history[index].type])
				self.targets[i]:SetText(self.history[index].target)
				self.timestamps[i]:SetText(os.date("%H:%M:%S", self.history[index].timestamp))
				if(index == self.selected) then
					self.entries[i]:SetBackgroundColor(0.3, 0.3, 0.3)
				else
					self.entries[i]:SetBackgroundColor(0, 0, 0)
				end
			end
		end
	end
	
	return self
end

local function createStatisticsList(parent, width, entries)
	local self = UI.CreateFrame("Frame", "", parent)
	self:SetBackgroundColor(0, 0, 0)
	self.selected = 0
	
	local scrollbar = UI.CreateFrame("RiftScrollbar", "", self)
	scrollbar:SetPoint("TOPLEFT", self, "TOPRIGHT")
	scrollbar:SetPoint("BOTTOMLEFT", self, "BOTTOMRIGHT")
	scrollbar.Event.ScrollbarChange = function() self:Update() end
	
	local prevAnchor = self
	self.entries = { }
	self.identifiers = { }
	self.types = { }
	self.sent = { }
	self.received = { }
	self.statistics = { }
	
	for i = 1, entries do
		local entry = UI.CreateFrame("Frame", "", self)
		entry:SetPoint("TOPLEFT", prevAnchor, prevAnchor == self and "TOPLEFT" or "BOTTOMLEFT")
		entry:SetPoint("TOPRIGHT", prevAnchor, prevAnchor == self and "TOPRIGHT" or "BOTTOMRIGHT")
		
		local received = UI.CreateFrame("Text", "", entry)
		received:SetPoint("TOPRIGHT", entry, "TOPRIGHT")
		received:SetText("99999")
		
		local sent = UI.CreateFrame("Text", "", entry)
		sent:SetPoint("TOPRIGHT", entry, "TOPRIGHT", -40, 0)
		sent:SetText("99999")
		
		local type = UI.CreateFrame("Text", "", entry)
		type:SetWidth(30)
		type:SetText("chan")
		type:SetPoint("TOPRIGHT", entry, "TOPRIGHT", -80, 0)
		
		local identifier = UI.CreateFrame("Text", "", entry)
		identifier:SetText("SomeIdentifier")
		identifier:SetPoint("TOPLEFT", entry, "TOPLEFT")
		identifier:SetPoint("TOPRIGHT", type, "TOPLEFT")
		
		entry:SetHeight(identifier:GetHeight())
		
		entry.Event.MouseIn = function(entry) entry:SetBackgroundColor(0.3, 0.3, 0.3) end
		entry.Event.MouseOut = function(entry)
			if(i + math.floor(scrollbar:GetPosition() + 0.5) ~= self.selected) then
				entry:SetBackgroundColor(0, 0, 0)
			end
		end
		entry.Event.LeftDown = function(entry)
			self.selected = i + math.floor(scrollbar:GetPosition() + 0.5)
			statisticSelected(parent, self, self.selected)
			self:Update()
		end
		table.insert(self.entries, entry)
		table.insert(self.identifiers, identifier)
		table.insert(self.types, type)
		table.insert(self.sent, sent)
		table.insert(self.received, received)
		prevAnchor = entry
	end
	
	self:SetHeight(self.entries[1]:GetHeight() * entries)
	self:SetWidth(width - scrollbar:GetWidth())
	
	local label = UI.CreateFrame("Text", "", self)
	label:SetText("recvd")
	label:SetPoint("BOTTOMRIGHT", self, "TOPRIGHT", 0, 3)
	
	local label = UI.CreateFrame("Text", "", self)
	label:SetText("sent")
	label:SetPoint("BOTTOMRIGHT", self, "TOPRIGHT", -40, 3)
	
	function self:Update(reorder)
		if(reorder) then
			self.statistics = { }
			self.selected = 0
			for ident, types in pairs(statistics) do
				for type, stats in pairs(types) do
					self.statistics[#self.statistics + 1] = { ident, type, stats }
				end
			end
			table.sort(self.statistics, function(a, b) return a[1] < b[1] end)
			scrollbar:SetRange(0, math.max(1, #self.statistics - entries))
			scrollbar:SetEnabled(#self.statistics > entries)
			scrollbar:SetPosition(0)
		end
		local offset = math.floor(scrollbar:GetPosition() + 0.5)
		for i = 1, entries do
			local index = i + offset
			if(index > #self.statistics) then
				self.entries[i]:SetVisible(false)
			else
				self.entries[i]:SetVisible(true)
				self.identifiers[i]:SetText(self.statistics[index][1])
				self.types[i]:SetText(typeShortcuts[self.statistics[index][2]])
				self.types[i]:SetFontColor(unpack(typeColors[self.statistics[index][2]]))
				self.sent[i]:SetText(tostring(self.statistics[index][3].sent))
				self.received[i]:SetText(tostring(self.statistics[index][3].received))
				if(index == self.selected) then
					self.entries[i]:SetBackgroundColor(0.3, 0.3, 0.3)
				else
					self.entries[i]:SetBackgroundColor(0, 0, 0)
				end
			end
		end
	end
	
	return self
end

local function createCompressionArea(parent, target, identifier, data, handler)
	local self = UI.CreateFrame("Frame", "", parent)
	self:SetBackgroundColor(0, 0, 0, 0.5)
	self:SetWidth(135)
	
	local label = UI.CreateFrame("Text", "", self)
	label:SetPoint("TOPLEFT", self, "TOPLEFT")
	label:SetFontSize(14)
	label:SetFontColor(1, 1, 0)
	label:SetText("Compression")
	
	local uncompressed = UI.CreateFrame("Text", "", self)
	uncompressed:SetText(string.format(labelFormats.uncompressed, 0))
	uncompressed:SetPoint("TOPLEFT", label, "BOTTOMLEFT", padding, 0)
	
	local compressed = UI.CreateFrame("Text", "", self)
	compressed:SetText(string.format(labelFormats.compressed, 0))
	compressed:SetPoint("TOPLEFT", uncompressed, "BOTTOMLEFT")
	
	local ratio = UI.CreateFrame("Text", "", self)
	ratio:SetText(string.format(labelFormats.ratio, 0))
	ratio:SetPoint("TOPLEFT", compressed, "BOTTOMLEFT")
	
	label = UI.CreateFrame("Text", "", self)
	label:SetPoint("TOPLEFT", ratio, "BOTTOMLEFT")
	label:SetText(string.format("Compr. Level (%i-%i)", zlib.BEST_SPEED, zlib.BEST_COMPRESSION))
	local level = createTextField(self, false)
	level:SetPoint("TOPRIGHT", ratio, "BOTTOMLEFT", 135 - padding - 1, 0)
	level:SetWidth(20)
	level:SetText("6")
	level.Event.KeyFocusLoss = function()
		local value = tonumber(level:GetText())
		if(value) then
			value = math.min(zlib.BEST_COMPRESSION, math.max(zlib.BEST_SPEED, value))
		else
			value = 6
		end
		level:SetText(tostring(value))
	end
	
	function self:Compress()
		local size1 = Utility.Message.Size(target(), identifier(), data())
		local data2 = zlib.deflate()(data(), "finish")
		local size2 = Utility.Message.Size(target(), identifier(), data2)
		uncompressed:SetText(string.format(labelFormats.uncompressed, size1))
		compressed:SetText(string.format(labelFormats.compressed, size2))
		ratio:SetText(string.format(labelFormats.ratio, math.floor(size2 / size1 * 100)))
		handler(data2)
	end

	function self:Uncompress()
		local size1 = Utility.Message.Size(target(), identifier(), data())
		local success, data2 = pcall(zlib.inflate(), data(), "finish")
		if(success) then
			local size2 = Utility.Message.Size(target(), identifier(), data2)
			compressed:SetText(string.format(labelFormats.compressed, size1))
			uncompressed:SetText(string.format(labelFormats.uncompressed, size2))
			ratio:SetText(string.format(labelFormats.ratio, math.floor(size1 / size2 * 100)))
			handler(data2)
		else
			print("Message is not in zlib format")
			print(data2)
		end
	end
	
	local deflateBtn = UI.CreateFrame("RiftButton", "", self)
	deflateBtn:SetText("Compress")
	deflateBtn:SetPoint("TOPLEFT", label, "BOTTOMLEFT", -padding, 0)
	deflateBtn.Event.LeftPress = self.Compress
	
	local inflateBtn = UI.CreateFrame("RiftButton", "", self)
	inflateBtn:SetText("Uncompress")
	inflateBtn:SetPoint("TOPLEFT", deflateBtn, "BOTTOMLEFT")
	inflateBtn.Event.LeftPress = self.Uncompress
	
	function self:Reset()
		compressed:SetText(string.format(labelFormats.compressed, 0))
		uncompressed:SetText(string.format(labelFormats.uncompressed, 0))
		ratio:SetText(string.format(labelFormats.ratio, 0))
	end
	
	function self:SetEnabled(enabled)
		deflateBtn:SetEnabled(enabled)
		inflateBtn:SetEnabled(enabled)
	end
	
	self:SetHeight(inflateBtn:GetBottom() - self:GetTop() + padding)
	return self
end

local function createMessageText(parent)
	local self = UI.CreateFrame("Frame", "", parent)
	
	local mask = UI.CreateFrame("Mask", "", parent)
	mask:SetBackgroundColor(0, 0, 0)
	
	local text = UI.CreateFrame("Text", "", mask)
	text:SetText(" ")
	text:SetWordwrap(true)
	text:SetPoint("TOPLEFT", mask, "TOPLEFT")
	text:SetPoint("TOPRIGHT", mask, "TOPRIGHT")
	
	local scrollbar = UI.CreateFrame("RiftScrollbar", "", parent)
	scrollbar:SetPoint("TOPLEFT", mask, "TOPRIGHT")
	scrollbar:SetPoint("BOTTOMLEFT", mask, "BOTTOMRIGHT")
	function scrollbar.Event.ScrollbarChange()
		local offset = scrollbar:GetPosition()
		text:SetPoint("TOPLEFT", mask, "TOPLEFT", 0, -offset)
		text:SetPoint("TOPRIGHT", mask, "TOPRIGHT", 0, -offset)
	end
	
	local function compressionHandler(data)
		self.data = data
		text:SetText(data)
	end
	local compression = createCompressionArea(self,
		function() return self.message.from end,
		function() return self.message.identifier end,
		function() return self.data end,
		compressionHandler
	)
	compression:SetPoint("TOPRIGHT", self, "TOPRIGHT")
	local autoUncompress = createCheckbox(compression, "Auto uncompress", false)
	autoUncompress:SetPoint("BOTTOMLEFT", compression, "BOTTOMLEFT", padding, -padding)
	compression:SetHeight(compression:GetHeight() + 15)

	mask:SetPoint("TOPLEFT", self, "TOPLEFT")
	mask:SetPoint("BOTTOMLEFT", self, "BOTTOMLEFT")
	mask:SetPoint("TOPRIGHT", compression, "TOPLEFT", -scrollbarWidth, 0)
	
	local time = UI.CreateFrame("Text", "", self)
	time:SetText(" ")
	time:SetPoint("BOTTOMRIGHT", mask, "TOPRIGHT")
	
	local size = UI.CreateFrame("Text", "", self)
	size:SetText(" ")
	size:SetPoint("BOTTOMRIGHT", time, "BOTTOMLEFT", -20, 0)
	
	function self:SetMessage(message)
		self.message = message
		if(message) then
			self.data = message.data
			text:SetText(message.data)
			text:SetFontColor(unpack(typeColors[message.type]))
			time:SetText(os.date("%H:%M:%S", message.timestamp))
			size:SetText(tostring(Utility.Message.Size(message.from, message.identifier, message.data)))
			if(message.from == "player") then
				time:SetFontColor(unpack(outgoingColor))
				size:SetFontColor(unpack(outgoingColor))
			else
				time:SetFontColor(unpack(incomingColor))
				size:SetFontColor(unpack(incomingColor))
			end
			if(autoUncompress:GetChecked()) then
				compression:Uncompress()
			end
		else
			text:SetText("")
			self.data = ""
		end
		compression:SetEnabled(message ~= nil)
		local h = text:GetHeight() - self:GetHeight()
		scrollbar:SetRange(0, math.max(1, h))
		scrollbar:SetEnabled(h > 0)
		scrollbar:SetPosition(0)
	end
	
	function self.Event.Size()
		local h = text:GetHeight() - self:GetHeight()
		scrollbar:SetRange(0, math.max(1, h))
		scrollbar:SetEnabled(h > 0)
		scrollbar:SetPosition(0)
	end
	
	return self
end

local function createNewMessageText(parent)
end

function acceptorSelected(self, list, index)
	self.historyList:SelectHistory(list.acceptors[index][1], list.acceptors[index][2])
	self.statisticsList.selected = 0
	self.statisticsList:Update()
end

function statisticSelected(self, list, index)
	self.historyList:SelectHistory(list.statistics[index][1], list.statistics[index][2])
	self.acceptorsList.selected = 0
	self.acceptorsList:Update()
end

function messageSelected(self, list, message)
	self.messageText:SetMessage(message)
end

local function createWindow()
	if(not Inspect.System.Secure()) then
		Command.System.Watchdog.Quiet()
	end
	
	local context = UI.CreateContext(Addon.identifier)
	local self = UI.CreateFrame("RiftWindow", "", context)
	self:SetTitle(Addon.toc.Name)
	self.data = ""
	
	local label
	
	self:SetController("content")
	self:SetWidth(935)
	self:SetHeight(565)
	self:SetPoint("TOPLEFT", UIParent, "TOPLEFT", (UIParent:GetWidth() - self:GetWidth()) / 2, (UIParent:GetHeight() - self:GetHeight()) / 2)
	
	local btn = UI.CreateFrame("RiftButton", "", self)
	btn:SetSkin("close")
	btn:SetPoint("TOPRIGHT", self:GetBorder(), "TOPRIGHT", -8, 16)
	btn.Event.LeftPress = function() self:SetVisible(false) end

	-- Statistics list
	label = UI.CreateFrame("Text", "", self)
	label:SetPoint("TOPLEFT", self, "TOPLEFT", padding, padding)
	label:SetText("Statistics")
	label:SetFontSize(16)
	
	self.statisticsList = createStatisticsList(self, identifierWidth + 110, 15)
	self.statisticsList:SetPoint("TOPLEFT", label, "BOTTOMLEFT")
	
	local btn = UI.CreateFrame("RiftButton", "", self)
	btn:SetText("Reset Counters")
	btn:SetPoint("BOTTOMRIGHT", self.statisticsList, "TOPRIGHT", scrollbarWidth, -12)
	btn.Event.LeftPress = function()
		resetCounters()
		self.statisticsList:Update(false)
	end
	
	-- List of active acceptors
	label = UI.CreateFrame("Text", "", self)
	label:SetPoint("TOPLEFT", self.statisticsList, "BOTTOMLEFT")
	label:SetText("Acceptors")
	label:SetFontSize(16)
	
	self.acceptorsList = createAcceptorsList(self, self.statisticsList:GetWidth(), 9)
	self.acceptorsList:SetPoint("TOPLEFT", label, "BOTTOMLEFT")

	-- List of messages
	label = UI.CreateFrame("Text", "", self)
	label:SetPoint("BOTTOMLEFT", self.statisticsList, "TOPRIGHT", scrollbarWidth, 0)
	label:SetText("Messages")
	label:SetFontSize(16)
	
	self.historyList = createHistoryList(self, 28)
	self.historyList:SetPoint("TOPLEFT", label, "BOTTOMLEFT")
	
	local btn = UI.CreateFrame("RiftButton", "", self)
	btn:SetText("Clear All")
	btn:SetPoint("BOTTOMRIGHT", self.historyList, "TOPRIGHT", -170, -12)
	btn.Event.LeftPress = clearEntireHistory
	
	local btn = UI.CreateFrame("RiftButton", "", self)
	btn:SetText("Clear List")
	btn:SetPoint("BOTTOMRIGHT", self.historyList, "TOPRIGHT", -45, -12)
	btn.Event.LeftPress = clearDisplayedHistory
	
	label = UI.CreateFrame("Text", "", self)
	label:SetText("incoming")
	label:SetFontColor(unpack(incomingColor))
	label:SetPoint("BOTTOMRIGHT", self.historyList, "TOPRIGHT", 0, 0)
	
	label = UI.CreateFrame("Text", "", self)
	label:SetText("outgoing")
	label:SetFontColor(unpack(outgoingColor))
	label:SetPoint("BOTTOMRIGHT", self.historyList, "TOPRIGHT", 0, -14)
	
	-- Message content area
	label = UI.CreateFrame("Text", "", self)
	label:SetText("Message Text")
	label:SetPoint("BOTTOMLEFT", self.historyList, "TOPRIGHT", scrollbarWidth, 0)
	label:SetFontSize(16)
	
	self.messageText = createMessageText(self)
	self.messageText:SetPoint("TOPLEFT", self.historyList, "TOPRIGHT", scrollbarWidth + 2, 0)
	self.messageText:SetWidth(self:GetRight() - self.messageText:GetLeft() - padding)
	self.messageText:SetHeight(200)
	
	-- Send message area
	label = UI.CreateFrame("Text", "", self)
	label:SetText("New Message")
	label:SetPoint("TOPLEFT", self.messageText, "BOTTOMLEFT")
	label:SetFontSize(16)
	
	self.newMessageText = createDataTextField(self)
	self.newMessageText:SetPoint("TOPLEFT", label, "BOTTOMLEFT")
	self.newMessageText:SetWidth(self.messageText:GetWidth() - 135 - padding)
	self.newMessageText:SetHeight(270)
	
	local size = UI.CreateFrame("Text", "", self)
	size:SetText(" ")
	size:SetPoint("BOTTOMRIGHT", self.newMessageText, "TOPRIGHT")
	
	local sendIdentifier = createTextField(self, "<identifier>")
	
	local sendTarget = createTextField(self, "<target>")
	function sendIdentifier.Event:KeyDown(button)
		if string.byte(button) == 9 then sendTarget:SetKeyFocus(true) end
	end
	function sendTarget.Event:KeyDown(button)
		if string.byte(button) == 9 then sendIdentifier:SetKeyFocus(true) end
	end
	
	function self.newMessageText.Event.TextfieldChange()
		self.data = self.newMessageText:GetText()
		size:SetText(tostring(Utility.Message.Size(sendTarget:GetText(), sendIdentifier:GetText(), self.data)))
	end
	
	local function handler(data)
		self.data = data
		self.newMessageText:SetText(data)
	end
	
	local compressor = createCompressionArea(self,
		function() return sendTarget:GetText() end,
		function() return sendIdentifier:GetText() end,
		function() return self.data end,
	handler)
	compressor:SetPoint("TOPLEFT", self.newMessageText, "TOPRIGHT", padding, 0)
	
	local autoCompress = createCheckbox(compressor, "Auto compress", false)
	autoCompress:SetPoint("BOTTOMLEFT", compressor, "BOTTOMLEFT", padding, -padding)
	compressor:SetHeight(compressor:GetHeight() + 15)

	local simulateIncoming = createCheckbox(self, "Incoming", false)
	simulateIncoming:SetPoint("LEFTCENTER", label, "RIGHTCENTER", padding, 0)
	
	local sendFrom = createTextField(self, "<sender>")
	sendFrom:SetPoint("BOTTOMLEFT", compressor, "TOPLEFT", 0, -2)
	sendFrom:SetPoint("BOTTOMRIGHT", compressor, "TOPRIGHT", 0, -2)
	sendFrom:GetParent():SetVisible(false)

	function simulateIncoming.Event.CheckboxChange()
		sendFrom:GetParent():SetVisible(simulateIncoming:GetChecked())
	end
	
	sendIdentifier:SetPoint("TOPLEFT", compressor, "BOTTOMLEFT", 0, padding)
	sendIdentifier:SetPoint("TOPRIGHT", compressor, "BOTTOMRIGHT", 0, padding)
	sendTarget:SetPoint("TOPLEFT", sendIdentifier, "BOTTOMLEFT", 0, padding)
	sendTarget:SetPoint("TOPRIGHT", sendIdentifier, "BOTTOMRIGHT", 0, padding)
	
	local types = { }
	types.party = createCheckbox(self, "party", true)
	types.party.label:SetFontColor(unpack(typeColors.party))
	types.party:SetPoint("TOPLEFT", self.newMessageText, "BOTTOMLEFT", padding, padding)
	
	types.raid = createCheckbox(self, "raid", false)
	types.raid.label:SetFontColor(unpack(typeColors.raid))
	types.raid:SetPoint("TOPLEFT", types.party, "TOPLEFT", 50, 0)
	
	types.guild = createCheckbox(self, "guild", false)
	types.guild.label:SetFontColor(unpack(typeColors.guild))
	types.guild:SetPoint("TOPLEFT", types.raid, "TOPLEFT", 50, 0)
	
	types.officer = createCheckbox(self, "officer", false)
	types.officer.label:SetFontColor(unpack(typeColors.officer))
	types.officer:SetPoint("TOPLEFT", types.guild, "TOPLEFT", 50, 0)
	
	types.tell = createCheckbox(self, "tell", false)
	types.tell.label:SetFontColor(unpack(typeColors.tell))
	types.tell:SetPoint("TOPLEFT", types.party, "BOTTOMLEFT", 0, 0)
	
	types.say = createCheckbox(self, "say", false)
	types.say.label:SetFontColor(unpack(typeColors.say))
	types.say:SetPoint("TOPLEFT", types.tell, "TOPLEFT", 50, 0)
	
	types.yell = createCheckbox(self, "yell", false)
	types.yell.label:SetFontColor(unpack(typeColors.yell))
	types.yell:SetPoint("TOPLEFT", types.say, "TOPLEFT", 50, 0)
	
	types.channel = createCheckbox(self, "channel", false)
	types.channel.label:SetFontColor(unpack(typeColors.channel))
	types.channel:SetPoint("TOPLEFT", types.yell, "TOPLEFT", 50, 0)
	
	local selectedType = "party"
	local typeChanging = false
	for k, cb in pairs(types) do
		function cb.Event.CheckboxChange()
			if(typeChanging) then return end
			typeChanging = true
			if(not cb:GetChecked()) then cb:SetChecked(true)
			else
				for k, v in pairs(types) do
					v:SetChecked(false)
				end
				cb:SetChecked(true)
				selectedType = cb.label:GetText()
			end
			typeChanging = false
		end
	end
	
	local sendBtn = UI.CreateFrame("RiftButton", "", self)
	sendBtn:SetPoint("TOPLEFT", sendTarget, "BOTTOMLEFT", 0, padding)
	sendBtn:SetPoint("TOPRIGHT", sendTarget, "BOTTOMRIGHT", 0, padding)
	sendBtn:SetText("Send")
	function sendBtn.Event.LeftPress()
		local callback = function(failure, message)
			if(failure) then
				print("Send failed:")
				print(message)
			end
		end
		if(autoCompress:GetChecked()) then
			compressor:Compress()
		end
		local target = sendTarget:GetText()
		local success, unit = pcall(Inspect.Unit.Detail, target)
		if(success and unit) then
			target = unit.name
		end
		if(simulateIncoming:GetChecked()) then
			simulateIncomingSend(sendFrom:GetText(), sendIdentifier:GetText(), self.data)
		else
			Command.Message.Send(target, sendIdentifier:GetText(), self.data, callback)
		end
		if(autoCompress:GetChecked()) then
			compressor:Uncompress()
		end
	end

	local broadcastBtn = UI.CreateFrame("RiftButton", "", self)
	broadcastBtn:SetPoint("TOPLEFT", sendBtn, "BOTTOMLEFT")
	broadcastBtn:SetPoint("TOPRIGHT", sendBtn, "BOTTOMRIGHT")
	broadcastBtn:SetText("Broadcast")
	function broadcastBtn.Event.LeftPress()
		local callback = function(failure, message)
			if(failure) then
				print("Broadcast failed:")
				print(message)
			end
		end
		if(autoCompress:GetChecked()) then
			compressor:Compress()
		end
		if(selectedType == "channel" or selectedType == "tell") then
			local target = sendTarget:GetText()
			if(selectedType == "tell") then
				local success, unit = pcall(Inspect.Unit.Detail, target)
				if(success and unit) then
					target = unit.name
				end
			end
			if(simulateIncoming:GetChecked()) then
				simulateIncomingBroadcast(sendFrom:GetText(), selectedType, target, sendIdentifier:GetText(), self.data)
			else
				Command.Message.Broadcast(selectedType, target, sendIdentifier:GetText(), self.data, callback)
			end
		else
			if(simulateIncoming:GetChecked()) then
				simulateIncomingBroadcast(sendFrom:GetText(), selectedType, nil, sendIdentifier:GetText(), self.data)
			else
				Command.Message.Broadcast(selectedType, nil, sendIdentifier:GetText(), self.data, callback)
			end
		end
		if(autoCompress:GetChecked()) then
			compressor:Uncompress()
		end
	end

	-- Add acceptor button/fields
	local newIdentifier = createTextField(self, "<identifier>")
	newIdentifier:SetPoint("TOPLEFT", self.acceptorsList, "BOTTOMLEFT", 0, padding)
	newIdentifier:SetWidth(135)
	
	local newType = createTextField(self, "<type>")
	newType:SetPoint("TOPLEFT", newIdentifier, "TOPRIGHT", padding, 0)
	newType:SetWidth(self.acceptorsList:GetRight() - newType:GetLeft() + scrollbarWidth - 4)
	function newType.Event:KeyDown(button)
		if string.byte(button) == 9 then newIdentifier:SetKeyFocus(true) end
	end
	function newIdentifier.Event:KeyDown(button)
		if string.byte(button) == 9 then newType:SetKeyFocus(true) end
	end
	
	local addAcceptorBtn = UI.CreateFrame("RiftButton", "", self)
	addAcceptorBtn:SetText("Add Acceptor")
	addAcceptorBtn:SetPoint("TOPLEFT", newIdentifier, "BOTTOMLEFT")
	addAcceptorBtn:SetEnabled(false)
	function addAcceptorBtn.Event.LeftPress()
		local identifier = newIdentifier:GetText()
		if(identifier == "*") then identifier = nil end
		local type = newType:GetText()
		if(type == "*") then type = nil end
		Command.Message.Accept(type, identifier)
	end
	
	function newIdentifier.Event.TextfieldChange()
		local ident = newIdentifier:GetText()
		local enable = ident == "*" or (#ident >= 3 and ident ~= "<identifier>")
		enable = enable and typeShortcuts[newType:GetText()] ~= nil
		addAcceptorBtn:SetEnabled(enable)
	end
	newType.Event.TextfieldChange = newIdentifier.Event.TextfieldChange
	
	-- Make the window movable
	self:GetContent().Event.LeftDown = function()
		local mouse = Inspect.Mouse()
		self.mouseOffsetX, self.mouseOffsetY = self:GetLeft() - mouse.x, self:GetTop() - mouse.y
	end
	self:GetContent().Event.LeftUp = function()
		self.mouseOffsetX, self.mouseOffsetY = nil, nil
	end
	self:GetContent().Event.MouseMove = function()
		if(self.mouseOffsetX) then
			local mouse = Inspect.Mouse()
			local x = mouse.x + self.mouseOffsetX
			local y = mouse.y + self.mouseOffsetY
			self:SetPoint("TOPLEFT", UIParent, "TOPLEFT", x, y)
		end
	end
	self:GetBorder().Event.LeftDown = self:GetContent().Event.LeftDown
	self:GetBorder().Event.LeftUp = self:GetContent().Event.LeftUp
	self:GetBorder().Event.MouseMove = self:GetContent().Event.MouseMove
	
	-- Populate with initial values
	self.acceptorsList:SetAcceptors(Inspect.Message.Accept.List())
	self.statisticsList:Update(true)
	self.historyList:SelectHistory("", "")
	
	MainWindow = self
end

local function historyChanged(self, identifier, type)
	if(not (identifier and type)) then
		self.historyList:UpdateHistory()
	elseif(self.historyList.identifier == "*" or self.historyList.identifier == identifier) then
		if(self.historyList.type == "*" or self.historyList.type == type) then
			self.historyList:UpdateHistory()
		end
	end
end

local function statisticsChanged(self, reorder)
	self.statisticsList:Update(reorder)
end

-- Message processing methods
-- ============================================================================

local function addToHistory(from, type, target, identifier, data)
	-- TODO: cleanup too old history entries
	messageHistory[#messageHistory + 1] = {
		from = from,
		type = type,
		target = target or "",
		data = data,
		timestamp = os.time(),
		identifier = identifier,
	}
	
	local reorder = false
	local t = statistics[identifier]
	if(not t) then
		reorder = true
		t = { }
	end
	statistics[identifier] = t
	
	local t2 = t[type]
	if(not t2) then
		reorder = true
		t2 = { sent = 0, received = 0 }
	end
	t[type] = t2
	if(from == "player") then
		t2.sent = t2.sent + 1
	else
		t2.received = t2.received + 1
	end
	
	if(MainWindow) then
		historyChanged(MainWindow, identifier, type)
		statisticsChanged(MainWindow, reorder)
	end
end

function getHistory(identifier, type)
	local history = { }

	for i = 1, #messageHistory do
		local message = messageHistory[i]
		if(identifier == "*" or message.identifier == identifier) then
			if(type == "*" or message.type == type) then
				history[#history + 1] = message
			end
		end
	end
	return history
end

function resetCounters()
	for ident, stats in pairs(statistics) do
		for type, stat in pairs(stats) do
			stat.sent = 0
			stat.received = 0
		end
	end
end

function clearEntireHistory()
	messageHistory = { }
	if(MainWindow) then
		historyChanged(MainWindow, nil, nil)
	end
end

function clearDisplayedHistory()
	local self = MainWindow
	
	local list = self.historyList.history
	if(#list <= 0) then
		return
	end
	
	local lookup = { }
	for i = 1, #list do
		lookup[list[i]] = true
	end
	
	local i = 1
	while(true) do
		local message = messageHistory[i]
		if(not message) then
			break
		end
		if(lookup[message]) then
			table.remove(messageHistory, i)
		else
			i = i + 1
		end
	end
	if(MainWindow) then
		historyChanged(self, self.historyList.identifier, self.historyList.type)
	end
end

local function dispatchSimulated(from, type, channel, identifier, data)
	if(Inspect.Message.Accept.Check(type, identifier)) then
		for i = 1, #Event.Message.Receive do
			local handler = Event.Message.Receive[i]
			Utility.Dispatch(function() handler[1](from, type, channel, identifier, data) end, handler[2], "event " .. handler[3])
		end
	else
		print(string.format("No acceptor for %s/%s configured", identifier, type))
	end
end

function simulateIncomingBroadcast(from, type, channel, identifier, data)
	dispatchSimulated(from, type, channel, identifier, data)
end

function simulateIncomingSend(from, identifier, data)
	dispatchSimulated(from, "send", nil, identifier, data)
end

function CommandMessageAccept(type, identifier)
	oldCommandMessageAccept(type, identifier)
	if(MainWindow) then
		MainWindow.acceptorsList:SetAcceptors(Inspect.Message.Accept.List())
	end
end
Command.Message.Accept = CommandMessageAccept

function CommandMessageBroadcast(type, target, identifier, data, callback)
	if(callback) then
		oldCommandMessageBroadcast(type, target, identifier, data, function(failure, message)
			if(failure) then
				print(message)
			else
				addToHistory("player", type, target, identifier, data)
			end
			callback(failure, message)
		end)
	else
		oldCommandMessageBroadcast(type, target, identifier, data)
		addToHistory("player", type, target, identifier, data)
	end
end
Command.Message.Broadcast = CommandMessageBroadcast

function CommandMessageReject(type, identifier)
	oldCommandMessageReject(type, identifier)
	if(MainWindow) then
		MainWindow.acceptorsList:SetAcceptors(Inspect.Message.Accept.List())
	end
end
Command.Message.Reject = CommandMessageReject

function CommandMessageSend(target, identifier, data, callback)
	oldCommandMessageSend(target, identifier, data, function(failure, message)
		if(failure) then
			print(message)
		else
			addToHistory("player", "send", target, identifier, data)
		end
		callback(failure, message)
	end)
end
Command.Message.Send = CommandMessageSend

local function messageReceived(from, type, channel, identifier, data)
	addToHistory(from, type, channel, identifier, data)
end

table.insert(Event.Message.Receive, { messageReceived, Addon.identifier, "messageReceived"})

local function toggleWindow()
	if(not MainWindow) then
		createWindow()
	else
		MainWindow:SetVisible(not MainWindow:GetVisible())
	end
end

table.insert(Command.Slash.Register(Addon.identifier), { toggleWindow, Addon.identifier, "toggleWindow" })
table.insert(Command.Slash.Register("mv"), { toggleWindow, Addon.identifier, "toggleWindow" })
Command.Message.Accept(nil, nil)
--toggleWindow()
--[[
Command.Message.Accept(nil, nil)
Command.Message.Accept(nil, nil)
Command.Message.Accept(nil,			"ImhoBags")
Command.Message.Accept("tell",		"ImhoBags")
Command.Message.Accept("channel",	"ImhoBags")
Command.Message.Accept("guild",		"ImhoBags")
Command.Message.Accept("officer",	"ImhoBags")
Command.Message.Accept("party",		"ImhoBags")
Command.Message.Accept("raid",		"ImhoBags")
Command.Message.Accept("say",		"ImhoBags")
Command.Message.Accept("yell",		"ImhoBags")
Command.Message.Accept("send",		"ImhoBags")
]]